@@ -1,26 +1,6 @@
-/* PCM - PipeWire plugin
- *
- * Copyright © 2017 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PCM - PipeWire plugin */
+/* SPDX-FileCopyrightText: Copyright © 2017 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #define __USE_GNU
 
@@ -39,20 +19,13 @@
 #include <spa/param/audio/format-utils.h>
 #include <spa/debug/types.h>
 #include <spa/param/props.h>
+#include <spa/utils/atomic.h>
 #include <spa/utils/result.h>
 #include <spa/utils/string.h>
+#include <spa/utils/json.h>
 
 #include <pipewire/pipewire.h>
 
-#define ATOMIC_INC(s)                   __atomic_add_fetch(&(s), 1, __ATOMIC_SEQ_CST)
-#define ATOMIC_LOAD(s)                  __atomic_load_n(&(s), __ATOMIC_SEQ_CST)
-
-#define SEQ_WRITE(s)                    ATOMIC_INC(s)
-#define SEQ_WRITE_SUCCESS(s1,s2)        ((s1) + 1 == (s2) && ((s2) & 1) == 0)
-
-#define SEQ_READ(s)                     ATOMIC_LOAD(s)
-#define SEQ_READ_SUCCESS(s1,s2)         ((s1) == (s2) && ((s2) & 1) == 0)
-
 PW_LOG_TOPIC_STATIC(alsa_log_topic, "alsa.pcm");
 #define PW_LOG_TOPIC_DEFAULT alsa_log_topic
 
@@ -70,30 +43,12 @@ PW_LOG_TOPIC_STATIC(alsa_log_topic, "alsa.pcm");
 #define MIN_BUFFER_BYTES	(2*MIN_PERIOD_BYTES)
 #define MAX_BUFFER_BYTES	(2*MAX_PERIOD_BYTES)
 
-struct params {
-	const char *node_name;
-	const char *server_name;
-	const char *playback_node;
-	const char *capture_node;
-	const char *role;
-	snd_pcm_format_t format;
-	int rate;
-	int channels;
-	int period_bytes;
-	int buffer_bytes;
-	uint32_t flags;
-};
-
 typedef struct {
 	snd_pcm_ioplug_t io;
 
 	snd_output_t *output;
 	FILE *log_file;
 
-	char *node_name;
-	char *target;
-	char *role;
-
 	int fd;
 	int error;
 	unsigned int activated:1;	/* PipeWire is activated? */
@@ -102,6 +57,7 @@ typedef struct {
 	unsigned int xrun_detected:1;
 	unsigned int hw_params_changed:1;
 	unsigned int active:1;
+	unsigned int negotiated:1;
 
 	snd_pcm_uframes_t hw_ptr;
 	snd_pcm_uframes_t boundary;
@@ -113,12 +69,12 @@ typedef struct {
 	struct spa_system *system;
 	struct pw_thread_loop *main_loop;
 
+	struct pw_properties *props;
 	struct pw_context *context;
 
 	struct pw_core *core;
 	struct spa_hook core_listener;
 
-	uint32_t flags;
 	struct pw_stream *stream;
 	struct spa_hook stream_listener;
 
@@ -192,9 +148,7 @@ static void snd_pcm_pipewire_free(snd_pcm_pipewire_t *pw)
 		spa_system_close(pw->system, pw->fd);
 	if (pw->main_loop)
 		pw_thread_loop_destroy(pw->main_loop);
-	free(pw->node_name);
-	free(pw->target);
-	free(pw->role);
+	pw_properties_free(pw->props);
 	snd_output_close(pw->output);
 	fclose(pw->log_file);
 	free(pw);
@@ -262,7 +216,7 @@ static int snd_pcm_pipewire_delay(snd_pcm_ioplug_t *io, snd_pcm_sframes_t *delay
 	int64_t diff;
 
 	do {
-		seq1 = SEQ_READ(pw->seq);
+		seq1 = SPA_SEQ_READ(pw->seq);
 
 		delay = pw->delay + pw->transfered;
 		now = pw->now;
@@ -271,8 +225,8 @@ static int snd_pcm_pipewire_delay(snd_pcm_ioplug_t *io, snd_pcm_sframes_t *delay
 		else
 			avail = snd_pcm_ioplug_avail(io, pw->hw_ptr, io->appl_ptr);
 
-		seq2 = SEQ_READ(pw->seq);
-	} while (!SEQ_READ_SUCCESS(seq1, seq2));
+		seq2 = SPA_SEQ_READ(pw->seq);
+	} while (!SPA_SEQ_READ_SUCCESS(seq1, seq2));
 
 	if (now != 0 && (io->state == SND_PCM_STATE_RUNNING ||
 	    io->state == SND_PCM_STATE_DRAINING)) {
@@ -425,6 +379,20 @@ static void on_stream_param_changed(void *data, uint32_t id, const struct spa_po
 			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(pw->stride));
 
 	pw_stream_update_params(pw->stream, params, n_params);
+
+	pw->negotiated = true;
+	pw_thread_loop_signal(pw->main_loop, false);
+}
+
+static void on_stream_state_changed(void *data, enum pw_stream_state old, enum pw_stream_state state, const char *error)
+{
+	snd_pcm_pipewire_t *pw = data;
+
+	if (state == PW_STREAM_STATE_ERROR) {
+		pw_log_warn("%s", error);
+		pw->error = -EIO;
+		update_active(&pw->io);
+	}
 }
 
 static void on_stream_drained(void *data)
@@ -462,7 +430,7 @@ static void on_stream_process(void *data)
 
 	want = b->requested ? b->requested : hw_avail;
 
-	SEQ_WRITE(pw->seq);
+	SPA_SEQ_WRITE(pw->seq);
 
 	if (pw->now != pwt.now) {
 		pw->transfered = pw->buffered;
@@ -480,7 +448,7 @@ static void on_stream_process(void *data)
 	pw->buffered = (want == 0 || pw->transfered < want) ?  0 : (pw->transfered % want);
 
 	pw->now = pwt.now;
-	SEQ_WRITE(pw->seq);
+	SPA_SEQ_WRITE(pw->seq);
 
 	pw_log_trace("%p: avail-before:%lu avail:%lu want:%lu xfer:%lu hw:%lu appl:%lu",
 			pw, before, hw_avail, want, xfer, pw->hw_ptr, io->appl_ptr);
@@ -503,6 +471,7 @@ done:
 static const struct pw_stream_events stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.param_changed = on_stream_param_changed,
+	.state_changed = on_stream_state_changed,
 	.process = on_stream_process,
 	.drained = on_stream_drained,
 };
@@ -541,7 +510,6 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 	const struct spa_pod *params[1];
 	uint8_t buffer[1024];
 	struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buffer, sizeof(buffer));
-	struct pw_properties *props;
 	uint32_t min_period;
 
 	pw_thread_loop_lock(pw->main_loop);
@@ -568,40 +536,18 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 		goto done;
 	pw->hw_params_changed = false;
 
-	props = pw_properties_new(NULL, NULL);
-	if (props == NULL)
-		goto error;
-
-	pw_properties_set(props, PW_KEY_CLIENT_API, "alsa");
-	pw_properties_setf(props, PW_KEY_APP_NAME, "%s", pw_get_prgname());
-
-	if (pw_properties_get(props, PW_KEY_NODE_LATENCY) == NULL)
-		pw_properties_setf(props, PW_KEY_NODE_LATENCY, "%lu/%u", pw->min_avail, io->rate);
-	if (pw_properties_get(props, PW_KEY_NODE_RATE) == NULL)
-		pw_properties_setf(props, PW_KEY_NODE_RATE, "1/%u", io->rate);
-	if (pw->target != NULL && !spa_streq(pw->target, "-1") &&
-	    pw_properties_get(props, PW_KEY_TARGET_OBJECT) == NULL)
-		pw_properties_setf(props, PW_KEY_TARGET_OBJECT, "%s", pw->target);
-
-	if (pw_properties_get(props, PW_KEY_MEDIA_TYPE) == NULL)
-		pw_properties_set(props, PW_KEY_MEDIA_TYPE, "Audio");
-	if (pw_properties_get(props, PW_KEY_MEDIA_CATEGORY) == NULL)
-		pw_properties_set(props, PW_KEY_MEDIA_CATEGORY,
-				io->stream == SND_PCM_STREAM_PLAYBACK ?
-				"Playback" : "Capture");
-	if (pw->role != NULL &&
-		pw_properties_get(props, PW_KEY_MEDIA_ROLE) == NULL)
-		pw_properties_setf(props, PW_KEY_MEDIA_ROLE, "%s", pw->role);
+	pw_properties_setf(pw->props, PW_KEY_NODE_LATENCY, "%lu/%u", pw->min_avail, io->rate);
+	pw_properties_setf(pw->props, PW_KEY_NODE_RATE, "1/%u", io->rate);
 
 	params[0] = spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &pw->format);
 
 	if (pw->stream != NULL) {
-		pw_stream_update_properties(pw->stream, &props->dict);
+		pw_stream_update_properties(pw->stream, &pw->props->dict);
 		pw_stream_update_params(pw->stream, params, 1);
 		goto done;
 	}
 
-	pw->stream = pw_stream_new(pw->core, pw->node_name, props);
+	pw->stream = pw_stream_new(pw->core, NULL, pw_properties_copy(pw->props));
 	if (pw->stream == NULL)
 		goto error;
 
@@ -609,12 +555,12 @@ static int snd_pcm_pipewire_prepare(snd_pcm_ioplug_t *io)
 
 	pw->error = 0;
 
+	pw->negotiated = false;
 	pw_stream_connect(pw->stream,
 				io->stream == SND_PCM_STREAM_PLAYBACK ?
 				PW_DIRECTION_OUTPUT :
 				PW_DIRECTION_INPUT,
 				PW_ID_ANY,
-				pw->flags |
 				PW_STREAM_FLAG_AUTOCONNECT |
 				PW_STREAM_FLAG_MAP_BUFFERS |
 				PW_STREAM_FLAG_RT_PROCESS,
@@ -627,13 +573,18 @@ done:
 	pw->drained = false;
 	pw->draining = false;
 
+	while (!pw->negotiated && pw->error >= 0)
+		pw_thread_loop_wait(pw->main_loop);
+	if (pw->error < 0)
+		goto error;
+
 	pw_thread_loop_unlock(pw->main_loop);
 
 	return 0;
 
 error:
 	pw_thread_loop_unlock(pw->main_loop);
-	return -ENOMEM;
+	return pw->error < 0 ? pw->error : -ENOMEM;
 }
 
 static int snd_pcm_pipewire_start(snd_pcm_ioplug_t *io)
@@ -969,108 +920,199 @@ static snd_pcm_ioplug_callback_t pipewire_pcm_callback = {
 	.query_chmaps = snd_pcm_pipewire_query_chmaps,
 };
 
-static int pipewire_set_hw_constraint(snd_pcm_pipewire_t *pw, struct params *p)
+#define MAX_VALS	64
+struct param_info {
+	const char *prop;
+	int key;
+#define TYPE_LIST	0
+#define TYPE_MIN_MAX	1
+	int type;
+	unsigned int vals[MAX_VALS];
+	unsigned int n_vals;
+	int (*collect) (const char *str, int len, unsigned int *val);
+
+};
+
+static int collect_access(const char *str, int len, unsigned int *val)
 {
-	unsigned int access_list[] = {
-		SND_PCM_ACCESS_MMAP_INTERLEAVED,
-		SND_PCM_ACCESS_MMAP_NONINTERLEAVED,
-		SND_PCM_ACCESS_RW_INTERLEAVED,
-		SND_PCM_ACCESS_RW_NONINTERLEAVED
-	};
-	unsigned int format_list[] = {
+	char key[64];
+
+	if (spa_json_parse_stringn(str, len, key, sizeof(key)) <= 0)
+		return -EINVAL;
+
+	if (strcasecmp(key, "MMAP_INTERLEAVED") == 0)
+		*val = SND_PCM_ACCESS_MMAP_INTERLEAVED;
+	else if (strcasecmp(key, "MMAP_NONINTERLEAVED") == 0)
+		*val = SND_PCM_ACCESS_MMAP_NONINTERLEAVED;
+	else if (strcasecmp(key, "RW_INTERLEAVED") == 0)
+		*val = SND_PCM_ACCESS_RW_INTERLEAVED;
+	else if (strcasecmp(key, "RW_NONINTERLEAVED") == 0)
+		*val = SND_PCM_ACCESS_RW_NONINTERLEAVED;
+	else
+		return -EINVAL;
+	return 0;
+}
+
+static int collect_format(const char *str, int len, unsigned int *val)
+{
+	char key[64];
+	snd_pcm_format_t fmt;
+
+	if (spa_json_parse_stringn(str, len, key, sizeof(key)) < 0)
+		return -EINVAL;
+
+	fmt = snd_pcm_format_value(key);
+	if (fmt != SND_PCM_FORMAT_UNKNOWN)
+		*val = fmt;
+	else
+		return -EINVAL;
+	return 0;
+}
+
+static int collect_int(const char *str, int len, unsigned int *val)
+{
+	int v;
+	if (spa_json_parse_int(str, len, &v) > 0)
+		*val = v;
+	else
+		return -EINVAL;
+	return 0;
+}
+
+struct param_info infos[] = {
+	{ "alsa.access", SND_PCM_IOPLUG_HW_ACCESS, TYPE_LIST,
+		{ SND_PCM_ACCESS_MMAP_INTERLEAVED,
+			SND_PCM_ACCESS_MMAP_NONINTERLEAVED,
+			SND_PCM_ACCESS_RW_INTERLEAVED,
+			SND_PCM_ACCESS_RW_NONINTERLEAVED }, 4, collect_access },
+	{ "alsa.format", SND_PCM_IOPLUG_HW_FORMAT, TYPE_LIST,
+		{
 #if __BYTE_ORDER == __LITTLE_ENDIAN
-		SND_PCM_FORMAT_FLOAT_LE,
-		SND_PCM_FORMAT_S32_LE,
-		SND_PCM_FORMAT_S24_LE,
-		SND_PCM_FORMAT_S24_3LE,
-		SND_PCM_FORMAT_S24_3BE,
-		SND_PCM_FORMAT_S16_LE,
+			SND_PCM_FORMAT_FLOAT_LE,
+			SND_PCM_FORMAT_S32_LE,
+			SND_PCM_FORMAT_S24_LE,
+			SND_PCM_FORMAT_S24_3LE,
+			SND_PCM_FORMAT_S24_3BE,
+			SND_PCM_FORMAT_S16_LE,
 #elif __BYTE_ORDER == __BIG_ENDIAN
-		SND_PCM_FORMAT_FLOAT_BE,
-		SND_PCM_FORMAT_S32_BE,
-		SND_PCM_FORMAT_S24_BE,
-		SND_PCM_FORMAT_S24_3LE,
-		SND_PCM_FORMAT_S24_3BE,
-		SND_PCM_FORMAT_S16_BE,
+			SND_PCM_FORMAT_FLOAT_BE,
+			SND_PCM_FORMAT_S32_BE,
+			SND_PCM_FORMAT_S24_BE,
+			SND_PCM_FORMAT_S24_3LE,
+			SND_PCM_FORMAT_S24_3BE,
+			SND_PCM_FORMAT_S16_BE,
 #endif
-		SND_PCM_FORMAT_U8,
-	};
-	int min_rate;
-	int max_rate;
-	int min_channels;
-	int max_channels;
-	int min_period_bytes;
-	int max_period_bytes;
-	int min_buffer_bytes;
-	int max_buffer_bytes;
+			SND_PCM_FORMAT_U8 }, 7, collect_format },
+	{ "alsa.rate", SND_PCM_IOPLUG_HW_RATE, TYPE_MIN_MAX,
+		{ 1, MAX_RATE }, 2, collect_int },
+	{ "alsa.channels", SND_PCM_IOPLUG_HW_CHANNELS, TYPE_MIN_MAX,
+		{ 1, MAX_CHANNELS }, 2, collect_int },
+	{ "alsa.buffer-bytes", SND_PCM_IOPLUG_HW_BUFFER_BYTES, TYPE_MIN_MAX,
+		{ MIN_BUFFER_BYTES, MAX_BUFFER_BYTES }, 2, collect_int },
+	{ "alsa.period-bytes", SND_PCM_IOPLUG_HW_PERIOD_BYTES, TYPE_MIN_MAX,
+		{ MIN_PERIOD_BYTES, MAX_PERIOD_BYTES }, 2, collect_int },
+	{ "alsa.periods", SND_PCM_IOPLUG_HW_PERIODS, TYPE_MIN_MAX,
+		{ MIN_BUFFERS, 1024 }, 2, collect_int },
+};
+
+static struct param_info *param_info_by_key(int key)
+{
+	SPA_FOR_EACH_ELEMENT_VAR(infos, p) {
+		if (p->key == key)
+			return p;
+	}
+	return NULL;
+}
+
+static int parse_value(const char *str, struct param_info *info)
+{
+	struct spa_json it[2];
+	unsigned int v;
+	const char *val;
+	int len;
+
+	spa_json_init(&it[0], str, strlen(str));
+	if ((len = spa_json_next(&it[0], &val)) <= 0)
+		return -EINVAL;
+
+	if (spa_json_is_array(val, len)) {
+		info->type = TYPE_LIST;
+		info->n_vals = 0;
+		spa_json_enter(&it[0], &it[1]);
+		while ((len = spa_json_next(&it[1], &val)) > 0 && info->n_vals < MAX_VALS) {
+			if (info->collect(val, len, &v) < 0)
+				continue;
+			info->vals[info->n_vals++] = v;
+		}
+	}
+	else if (spa_json_is_object(val, len)) {
+		char key[64];
+		info->type = TYPE_MIN_MAX;
+		info->n_vals = 2;
+		spa_json_enter(&it[0], &it[1]);
+                while (spa_json_get_string(&it[1], key, sizeof(key)) > 0) {
+                        if ((len = spa_json_next(&it[1], &val)) <= 0)
+                                break;
+			if (info->collect(val, len, &v) < 0)
+				continue;
+			if (spa_streq(key, "min"))
+				info->vals[0] = v;
+			else if (spa_streq(key, "max"))
+				info->vals[1] = v;
+                }
+	}
+	else if (info->collect(val, len, &v) >= 0) {
+		info->type = TYPE_LIST;
+		info->vals[0] = v;
+		info->n_vals = 1;
+	}
+	return 0;
+}
+
+static int set_constraint(snd_pcm_pipewire_t *pw, int key)
+{
+	struct param_info *p = param_info_by_key(key), info;
+	const char *str;
 	int err;
 
-	if (p->rate > 0) {
-		min_rate = max_rate = SPA_CLAMP(p->rate, 1, MAX_RATE);
-	} else {
-		min_rate = 1;
-		max_rate = MAX_RATE;
-	}
-	if (p->channels > 0) {
-		min_channels = max_channels = SPA_CLAMP(p->channels, 1, MAX_CHANNELS);
-	} else {
-		min_channels = 1;
-		max_channels = MAX_CHANNELS;
-	}
-	if (p->period_bytes > 0) {
-		min_period_bytes = max_period_bytes = SPA_CLAMP(p->period_bytes,
-				MIN_PERIOD_BYTES, MAX_PERIOD_BYTES);
-	} else {
-		min_period_bytes = MIN_PERIOD_BYTES;
-		max_period_bytes = MAX_PERIOD_BYTES;
-	}
-	if (p->buffer_bytes > 0) {
-		min_buffer_bytes = max_buffer_bytes = SPA_CLAMP(p->buffer_bytes,
-				MIN_BUFFER_BYTES, MAX_BUFFER_BYTES);
-	} else {
-		min_buffer_bytes = MIN_BUFFER_BYTES;
-		max_buffer_bytes = MAX_BUFFER_BYTES;
-	}
-	if (min_period_bytes * 2 > max_buffer_bytes)
-		min_period_bytes = max_period_bytes = max_buffer_bytes / 2;
+	if (p == NULL)
+		return -EINVAL;
 
-	if ((err = snd_pcm_ioplug_set_param_list(&pw->io, SND_PCM_IOPLUG_HW_ACCESS,
-						   SPA_N_ELEMENTS(access_list), access_list)) < 0 ||
-		(err = snd_pcm_ioplug_set_param_minmax(&pw->io, SND_PCM_IOPLUG_HW_CHANNELS,
-						   min_channels, max_channels)) < 0 ||
-		(err = snd_pcm_ioplug_set_param_minmax(&pw->io, SND_PCM_IOPLUG_HW_RATE,
-						   min_rate, max_rate)) < 0 ||
-		(err = snd_pcm_ioplug_set_param_minmax(&pw->io, SND_PCM_IOPLUG_HW_BUFFER_BYTES,
-						   min_buffer_bytes,
-						   max_buffer_bytes)) < 0 ||
-		(err = snd_pcm_ioplug_set_param_minmax(&pw->io,
-						   SND_PCM_IOPLUG_HW_PERIOD_BYTES,
-						   min_period_bytes,
-						   max_period_bytes)) < 0 ||
-		(err = snd_pcm_ioplug_set_param_minmax(&pw->io, SND_PCM_IOPLUG_HW_PERIODS,
-						   MIN_BUFFERS, 1024)) < 0) {
-		pw_log_warn("Can't set param list: %s", snd_strerror(err));
+	info = *p;
+
+	str = pw_properties_get(pw->props, p->prop);
+	if (str != NULL)
+		parse_value(str, &info);
+
+	switch (info.type) {
+	case TYPE_LIST:
+		pw_log_info("%s: list %d", p->prop, info.n_vals);
+		err = snd_pcm_ioplug_set_param_list(&pw->io, key, info.n_vals, info.vals);
+		break;
+	case TYPE_MIN_MAX:
+		pw_log_info("%s: min:%u max:%u", p->prop, info.vals[0], info.vals[1]);
+		err = snd_pcm_ioplug_set_param_minmax(&pw->io, key, info.vals[0], info.vals[1]);
+		break;
+	default:
+		return -EIO;
+	}
+	if (err < 0)
+		pw_log_warn("Can't set param %s: %s", info.prop, snd_strerror(err));
+
+	return err;
+
+}
+static int pipewire_set_hw_constraint(snd_pcm_pipewire_t *pw)
+{
+	int err;
+	if ((err = set_constraint(pw, SND_PCM_IOPLUG_HW_ACCESS)) < 0 ||
+	    (err = set_constraint(pw, SND_PCM_IOPLUG_HW_FORMAT)) < 0 ||
+	    (err = set_constraint(pw, SND_PCM_IOPLUG_HW_RATE)) < 0 ||
+	    (err = set_constraint(pw, SND_PCM_IOPLUG_HW_CHANNELS)) < 0 ||
+	    (err = set_constraint(pw, SND_PCM_IOPLUG_HW_PERIOD_BYTES)) < 0 ||
+	    (err = set_constraint(pw, SND_PCM_IOPLUG_HW_BUFFER_BYTES)) < 0 ||
+	    (err = set_constraint(pw, SND_PCM_IOPLUG_HW_PERIODS)) < 0)
 		return err;
-	}
-
-	if (p->format != SND_PCM_FORMAT_UNKNOWN) {
-		err = snd_pcm_ioplug_set_param_list(&pw->io,
-				SND_PCM_IOPLUG_HW_FORMAT,
-				1, (unsigned int *)&p->format);
-		if (err < 0) {
-			pw_log_warn("Can't set param list: %s", snd_strerror(err));
-			return err;
-		}
-	} else {
-		err = snd_pcm_ioplug_set_param_list(&pw->io,
-				SND_PCM_IOPLUG_HW_FORMAT,
-				SPA_N_ELEMENTS(format_list),
-				format_list);
-		if (err < 0) {
-			pw_log_warn("Can't set param list: %s", snd_strerror(err));
-			return err;
-		}
-	}
 	return 0;
 }
 
@@ -1113,61 +1155,31 @@ static cookie_io_functions_t io_funcs = {
 	.write = log_write,
 };
 
+static int execute_match(void *data, const char *location, const char *action,
+                const char *val, size_t len)
+{
+	snd_pcm_pipewire_t *pw = data;
+	if (spa_streq(action, "update-props"))
+		pw_properties_update_string(pw->props, val, len);
+	return 1;
+}
+
 static int snd_pcm_pipewire_open(snd_pcm_t **pcmp,
-		struct params *p, snd_pcm_stream_t stream, int mode)
+		struct pw_properties *props, snd_pcm_stream_t stream, int mode)
 {
 	snd_pcm_pipewire_t *pw;
 	int err;
-	const char *str;
-	struct pw_properties *props = NULL;
+	const char *str, *node_name = NULL;
 	struct pw_loop *loop;
-	uint32_t val;
 
 	assert(pcmp);
 	pw = calloc(1, sizeof(*pw));
 	if (!pw)
 		return -ENOMEM;
 
-	props = pw_properties_new(NULL, NULL);
-	if (props == NULL) {
-		err = -errno;
-		goto error;
-	}
-
-	str = getenv("PIPEWIRE_ALSA");
-	if (str != NULL) {
-		pw_properties_update_string(props, str, strlen(str));
-		if ((str = pw_properties_get(props, "alsa.format")))
-			p->format = snd_pcm_format_value(str);
-		if ((str = pw_properties_get(props, "alsa.rate")) &&
-		    spa_atou32(str, &val, 0))
-			p->rate = val;
-		if ((str = pw_properties_get(props, "alsa.channels")) &&
-		    spa_atou32(str, &val, 0))
-			p->channels = val;
-		if ((str = pw_properties_get(props, "alsa.period-bytes")) &&
-		    spa_atou32(str, &val, 0))
-			p->period_bytes = val;
-		if ((str = pw_properties_get(props, "alsa.buffer-bytes")) &&
-		    spa_atou32(str, &val, 0))
-			p->buffer_bytes = val;
-	}
-
-	str = getenv("PIPEWIRE_REMOTE");
-	if (str != NULL && str[0] != '\0')
-		p->server_name = str;
-
-	str = getenv("PIPEWIRE_NODE");
-
-	pw_log_debug("%p: open name:%s stream:%s mode:%d flags:%08x rate:%d format:%s "
-			"channels:%d period-bytes:%d buffer-bytes:%d target:'%s'", pw, p->node_name,
-			snd_pcm_stream_name(stream), mode, p->flags, p->rate,
-			snd_pcm_format_name(p->format), p->channels, p->period_bytes,
-			p->buffer_bytes, str);
-
+	pw->props = props;
 	pw->fd = -1;
 	pw->io.poll_fd = -1;
-	pw->flags = p->flags;
 	pw->log_file = fopencookie(pw, "w", io_funcs);
 	if (pw->log_file == NULL) {
 		pw_log_error("can't create log file: %m");
@@ -1179,29 +1191,6 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp,
 		goto error;
 	}
 
-	if (p->node_name == NULL)
-		pw->node_name = spa_aprintf("ALSA %s",
-			       stream == SND_PCM_STREAM_PLAYBACK ? "Playback" : "Capture");
-	else
-		pw->node_name = strdup(p->node_name);
-
-	if (pw->node_name == NULL) {
-		err = -errno;
-		goto error;
-	}
-
-	pw->target = NULL;
-	if (str != NULL)
-		pw->target = strdup(str);
-	else {
-		if (stream == SND_PCM_STREAM_PLAYBACK)
-			pw->target = p->playback_node ? strdup(p->playback_node) : NULL;
-		else
-			pw->target = p->capture_node ? strdup(p->capture_node) : NULL;
-	}
-
-	pw->role = (p->role && *p->role) ? strdup(p->role) : NULL;
-
 	pw->main_loop = pw_thread_loop_new("alsa-pipewire", NULL);
 	if (pw->main_loop == NULL) {
 		err = -errno;
@@ -1212,24 +1201,56 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp,
 	if ((pw->context = pw_context_new(loop,
 					pw_properties_new(
 						PW_KEY_CONFIG_NAME, "client-rt.conf",
+						PW_KEY_CLIENT_API, "alsa",
 						NULL),
 					0)) == NULL) {
 		err = -errno;
 		goto error;
 	}
 
-	pw_properties_setf(props, PW_KEY_APP_NAME, "PipeWire ALSA [%s]",
-			pw_get_prgname());
+	pw_context_conf_update_props(pw->context, "alsa.properties", pw->props);
 
-	if (p->server_name)
-		pw_properties_set(props, PW_KEY_REMOTE_NAME, p->server_name);
+	pw_context_conf_section_match_rules(pw->context, "alsa.rules",
+			&pw_context_get_properties(pw->context)->dict, execute_match, pw);
+
+	if (pw_properties_get(pw->props, PW_KEY_APP_NAME) == NULL)
+		pw_properties_setf(pw->props, PW_KEY_APP_NAME, "PipeWire ALSA [%s]",
+				pw_get_prgname());
+	if (pw_properties_get(pw->props, PW_KEY_NODE_NAME) == NULL)
+		pw_properties_setf(pw->props, PW_KEY_NODE_NAME, "alsa_%s.%s",
+			       stream == SND_PCM_STREAM_PLAYBACK ? "playback" : "capture",
+			       pw_get_prgname());
+	if (pw_properties_get(pw->props, PW_KEY_NODE_DESCRIPTION) == NULL)
+		pw_properties_setf(pw->props, PW_KEY_NODE_DESCRIPTION, "ALSA %s [%s]",
+			       stream == SND_PCM_STREAM_PLAYBACK ? "Playback" : "Capture",
+			       pw_get_prgname());
+	if (pw_properties_get(pw->props, PW_KEY_MEDIA_NAME) == NULL)
+		pw_properties_setf(pw->props, PW_KEY_MEDIA_NAME, "ALSA %s",
+			       stream == SND_PCM_STREAM_PLAYBACK ? "Playback" : "Capture");
+	if (pw_properties_get(pw->props, PW_KEY_MEDIA_TYPE) == NULL)
+		pw_properties_set(pw->props, PW_KEY_MEDIA_TYPE, "Audio");
+	if (pw_properties_get(pw->props, PW_KEY_MEDIA_CATEGORY) == NULL)
+		pw_properties_set(pw->props, PW_KEY_MEDIA_CATEGORY,
+				stream == SND_PCM_STREAM_PLAYBACK ?
+				"Playback" : "Capture");
+
+	str = getenv("PIPEWIRE_ALSA");
+	if (str != NULL)
+		pw_properties_update_string(pw->props, str, strlen(str));
+
+	str = getenv("PIPEWIRE_NODE");
+	if (str != NULL && str[0])
+		pw_properties_set(pw->props, PW_KEY_TARGET_OBJECT, str);
+
+	node_name = pw_properties_get(pw->props, PW_KEY_NODE_NAME);
+	if (pw_properties_get(pw->props, PW_KEY_MEDIA_NAME) == NULL)
+		pw_properties_set(pw->props, PW_KEY_MEDIA_NAME, node_name);
 
 	if ((err = pw_thread_loop_start(pw->main_loop)) < 0)
 		goto error;
 
 	pw_thread_loop_lock(pw->main_loop);
-	pw->core = pw_context_connect(pw->context, props, 0);
-	props = NULL;
+	pw->core = pw_context_connect(pw->context, pw_properties_copy(pw->props), 0);
 	if (pw->core == NULL) {
 		err = -errno;
 		pw_thread_loop_unlock(pw->main_loop);
@@ -1254,13 +1275,13 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp,
 #endif
 	pw->io.flags |= SND_PCM_IOPLUG_FLAG_MONOTONIC;
 
-	if ((err = snd_pcm_ioplug_create(&pw->io, p->node_name, stream, mode)) < 0)
+	if ((err = snd_pcm_ioplug_create(&pw->io, node_name, stream, mode)) < 0)
 		goto error;
 
-	if ((err = pipewire_set_hw_constraint(pw, p)) < 0)
+	if ((err = pipewire_set_hw_constraint(pw)) < 0)
 		goto error;
 
-	pw_log_debug("%p: opened name:%s stream:%s mode:%d", pw, p->node_name,
+	pw_log_debug("%p: opened name:%s stream:%s mode:%d", pw, node_name,
 			snd_pcm_stream_name(pw->io.stream), mode);
 
 	*pcmp = pw->io.pcm;
@@ -1268,8 +1289,7 @@ static int snd_pcm_pipewire_open(snd_pcm_t **pcmp,
 	return 0;
 
 error:
-	pw_log_debug("%p: failed to open %s :%s", pw, p->node_name, spa_strerror(err));
-	pw_properties_free(props);
+	pw_log_debug("%p: failed to open %s :%s", pw, node_name, spa_strerror(err));
 	snd_pcm_pipewire_free(pw);
 	return err;
 }
@@ -1279,16 +1299,20 @@ SPA_EXPORT
 SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 {
 	snd_config_iterator_t i, next;
-	struct params params;
+	struct pw_properties *props;
+	const char *str;
+	long val;
 	int err;
 
 	pw_init(NULL, NULL);
 	if (strstr(pw_get_library_version(), "0.2") != NULL)
 		return -ENOTSUP;
 
+	props = pw_properties_new(NULL, NULL);
+	if (props == NULL)
+		return -errno;
+
 	PW_LOG_TOPIC_INIT(alsa_log_topic);
-	spa_zero(params);
-	params.format = SND_PCM_FORMAT_UNKNOWN;
 
 	snd_config_for_each(i, next, conf) {
 		snd_config_t *n = snd_config_iterator_entry(i);
@@ -1298,83 +1322,93 @@ SND_PCM_PLUGIN_DEFINE_FUNC(pipewire)
 		if (spa_streq(id, "comment") || spa_streq(id, "type") || spa_streq(id, "hint"))
 			continue;
 		if (spa_streq(id, "name")) {
-			snd_config_get_string(n, &params.node_name);
+			if (snd_config_get_string(n, &str) == 0)
+				pw_properties_set(props, PW_KEY_NODE_NAME, str);
 			continue;
 		}
 		if (spa_streq(id, "server")) {
-			snd_config_get_string(n, &params.server_name);
+			if (snd_config_get_string(n, &str) == 0)
+				pw_properties_set(props, PW_KEY_REMOTE_NAME, str);
 			continue;
 		}
 		if (spa_streq(id, "playback_node")) {
-			snd_config_get_string(n, &params.playback_node);
+			if (stream == SND_PCM_STREAM_PLAYBACK &&
+			    snd_config_get_string(n, &str) == 0)
+				if (str != NULL && !spa_streq(str, "-1"))
+					pw_properties_set(props, PW_KEY_TARGET_OBJECT, str);
 			continue;
 		}
 		if (spa_streq(id, "capture_node")) {
-			snd_config_get_string(n, &params.capture_node);
+			if (stream == SND_PCM_STREAM_CAPTURE &&
+			    snd_config_get_string(n, &str) == 0)
+				if (str != NULL && !spa_streq(str, "-1"))
+					pw_properties_set(props, PW_KEY_TARGET_OBJECT, str);
 			continue;
 		}
 		if (spa_streq(id, "role")) {
-			snd_config_get_string(n, &params.role);
+			if (snd_config_get_string(n, &str) == 0)
+				if (str != NULL && *str)
+					pw_properties_set(props, PW_KEY_MEDIA_ROLE, str);
 			continue;
 		}
 		if (spa_streq(id, "exclusive")) {
 			if (snd_config_get_bool(n))
-				params.flags |= PW_STREAM_FLAG_EXCLUSIVE;
+				pw_properties_set(props, PW_KEY_NODE_EXCLUSIVE, "true");
 			continue;
 		}
 		if (spa_streq(id, "rate")) {
-			long val;
-
-			if (snd_config_get_integer(n, &val) == 0)
-				params.rate = val;
-			else
+			if (snd_config_get_integer(n, &val) == 0) {
+				if (val != 0)
+					pw_properties_setf(props, "alsa.rate", "%ld", val);
+			} else {
 				SNDERR("%s: invalid type", id);
+			}
 			continue;
 		}
 		if (spa_streq(id, "format")) {
-			const char *str;
-
 			if (snd_config_get_string(n, &str) == 0) {
-				params.format = snd_pcm_format_value(str);
-				if (*str && params.format == SND_PCM_FORMAT_UNKNOWN)
-					SNDERR("%s: invalid value %s", id, str);
+				if (str != NULL && *str)
+					pw_properties_set(props, "alsa.format", str);
 			} else {
 				SNDERR("%s: invalid type", id);
 			}
 			continue;
 		}
 		if (spa_streq(id, "channels")) {
-			long val;
-
-			if (snd_config_get_integer(n, &val) == 0)
-				params.channels = val;
-			else
+			if (snd_config_get_integer(n, &val) == 0) {
+				if (val != 0)
+					pw_properties_setf(props, "alsa.channels", "%ld", val);
+			} else {
 				SNDERR("%s: invalid type", id);
+			}
 			continue;
 		}
 		if (spa_streq(id, "period_bytes")) {
-			long val;
-
-			if (snd_config_get_integer(n, &val) == 0)
-				params.period_bytes = val;
-			else
+			if (snd_config_get_integer(n, &val) == 0) {
+				if (val != 0)
+					pw_properties_setf(props, "alsa.period-bytes", "%ld", val);
+			} else {
 				SNDERR("%s: invalid type", id);
+			}
 			continue;
 		}
 		if (spa_streq(id, "buffer_bytes")) {
 			long val;
 
-			if (snd_config_get_integer(n, &val) == 0)
-				params.buffer_bytes = val;
-			else
+			if (snd_config_get_integer(n, &val) == 0) {
+				if (val != 0)
+					pw_properties_setf(props, "alsa.buffer-bytes", "%ld", val);
+			} else {
 				SNDERR("%s: invalid type", id);
+			}
 			continue;
 		}
 		SNDERR("Unknown field %s", id);
+		pw_properties_free(props);
 		return -EINVAL;
 	}
 
-	err = snd_pcm_pipewire_open(pcmp, &params, stream, mode);
+	err = snd_pcm_pipewire_open(pcmp, props, stream, mode);
 
 	return err;
 }
