@@ -20,6 +20,7 @@
 #include <spa/utils/json.h>
 #include <spa/support/cpu.h>
 #include <spa/param/latency-utils.h>
+#include <spa/param/tag-utils.h>
 #include <spa/pod/dynamic.h>
 #include <spa/debug/types.h>
 
@@ -81,6 +82,12 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
  *         ]
  *         inputs = [ <portname> ... ]
  *         outputs = [ <portname> ... ]
+ *         capture.volumes = [
+ *             { control = <portname>  min = <value>  max = <value>  scale = <scale> } ...
+ *         ]
+ *         playback.volumes = [
+ *             { control = <portname>  min = <value>  max = <value>  scale = <scale> } ...
+ *         ]
  *    }
  *\endcode
  *
@@ -136,6 +143,20 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
  * graph will then be duplicated as many times to match the number of input/output
  * channels of the streams.
  *
+ * ### Volumes
+ *
+ * Normally the volume of the sink/source is handled by the stream software volume.
+ * With the capture.volumes and playback.volumes properties this can be handled
+ * by a control port in the graph instead.
+ *
+ * The min and max values (defaults 0.0 and 1.0) respectively can be used to scale
+ * and translate the volume min and max values.
+ *
+ * Normally the control values are linear and it is assumed that the plugin does not
+ * perform any scaling to the values. This can be changed with the scale property. By
+ * default this is linear but it can be set to cubic when the control applies a
+ * cubic transformation.
+ *
  * ## Builtin filters
  *
  * There are some useful builtin filters available. You select them with the label
@@ -147,7 +168,7 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
  *
  * The mixer plugin has up to 8 input ports labeled "In 1" to "In 8" and each with
  * a gain control labeled "Gain 1" to "Gain 8". There is an output port labeled
- * "Out". Unused input ports will be ignoded and not cause overhead.
+ * "Out". Unused input ports will be ignored and not cause overhead.
  *
  * ### Copy
  *
@@ -162,7 +183,14 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
  *
  * All biquad filters have an input port "In" and an output port "Out". They have
  * a "Freq", "Q" and "Gain" control. Their meaning depends on the particular biquad that
- * is used. The following labels can be used:
+ * is used. The biquads also have "b0", "b1", "b2", "a0", "a1" and "a2" ports that
+ * are read-only except for the bq_raw biquad, which can configure default values
+ * depending on the graph rate and change those at runtime.
+ *
+ * We refer to https://arachnoid.com/BiQuadDesigner/index.html for an explanation of
+ * the controls.
+ *
+ * The following labels can be used:
  *
  * - `bq_lowpass` a lowpass filter.
  * - `bq_highpass` a highpass filter.
@@ -172,6 +200,30 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
  * - `bq_peaking` a peaking filter.
  * - `bq_notch` a notch filter.
  * - `bq_allpass` an allpass filter.
+ * - `bq_raw` a raw biquad filter. You need a config section to specify coefficients
+ *   		per sample rate. The coefficients of the sample rate closest to the
+ *   		graph rate are selected:
+ *
+ *\code{.unparsed}
+ * filter.graph = {
+ *     nodes = [
+ *         {
+ *             type   = builtin
+ *             name   = ...
+ *             label  = bq_raw
+ *             config = {
+ *                 coefficients = [
+ *                     { rate =  44100, b0=.., b1=.., b2=.., a0=.., a1=.., a2=.. },
+ *                     { rate =  48000, b0=.., b1=.., b2=.., a0=.., a1=.., a2=.. },
+ *                     { rate = 192000, b0=.., b1=.., b2=.., a0=.., a1=.., a2=.. }
+ *                 ]
+ *             }
+ *             ...
+ *         }
+ *     }
+ *     ...
+ * }
+ *\endcode
  *
  * ### Convolver
  *
@@ -487,9 +539,16 @@ static const struct spa_dict_item module_props[] = {
 #define MAX_HNDL 64
 #define MAX_SAMPLES 8192
 
+#define DEFAULT_RATE	48000
+
 static float silence_data[MAX_SAMPLES];
 static float discard_data[MAX_SAMPLES];
 
+struct fc_plugin *load_ladspa_plugin(const struct spa_support *support, uint32_t n_support,
+		struct dsp_ops *dsp, const char *path, const char *config);
+struct fc_plugin *load_builtin_plugin(const struct spa_support *support, uint32_t n_support,
+		struct dsp_ops *dsp, const char *path, const char *config);
+
 struct plugin {
 	struct spa_list link;
 	int ref;
@@ -500,6 +559,13 @@ struct plugin {
 	struct spa_list descriptor_list;
 };
 
+struct plugin_func {
+	struct spa_list link;
+	char type[256];
+	fc_plugin_load_func *func;
+	void *hndl;
+};
+
 struct descriptor {
 	struct spa_list link;
 	int ref;
@@ -530,7 +596,7 @@ struct port {
 	uint32_t n_links;
 	uint32_t external;
 
-	float control_data;
+	float control_data[MAX_HNDL];
 	float *audio_data[MAX_HNDL];
 };
 
@@ -579,6 +645,20 @@ struct graph_hndl {
 	void **hndl;
 };
 
+struct volume {
+	bool mute;
+	uint32_t n_volumes;
+	float volumes[SPA_AUDIO_MAX_CHANNELS];
+
+	uint32_t n_ports;
+	struct port *ports[SPA_AUDIO_MAX_CHANNELS];
+	float min[SPA_AUDIO_MAX_CHANNELS];
+	float max[SPA_AUDIO_MAX_CHANNELS];
+#define SCALE_LINEAR	0
+#define SCALE_CUBIC	1
+	int scale[SPA_AUDIO_MAX_CHANNELS];
+};
+
 struct graph {
 	struct impl *impl;
 
@@ -596,6 +676,11 @@ struct graph {
 
 	uint32_t n_control;
 	struct port **control_port;
+
+	struct volume capture_volume;
+	struct volume playback_volume;
+
+	unsigned instantiated:1;
 };
 
 struct impl {
@@ -612,6 +697,7 @@ struct impl {
 	struct dsp_ops dsp;
 
 	struct spa_list plugin_list;
+	struct spa_list plugin_func_list;
 
 	struct pw_properties *capture_props;
 	struct pw_stream *capture;
@@ -623,6 +709,10 @@ struct impl {
 	struct spa_hook playback_listener;
 	struct spa_audio_info_raw playback_info;
 
+	struct spa_audio_info_raw info;
+
+	struct spa_io_position *position;
+
 	unsigned int do_disconnect:1;
 
 	long unsigned rate;
@@ -657,7 +747,16 @@ static void playback_process(void *d)
 	struct graph_port *port;
 	struct spa_data *bd;
 
-	if ((in = pw_stream_dequeue_buffer(impl->capture)) == NULL)
+	in = NULL;
+	while (true) {
+		struct pw_buffer *t;
+		if ((t = pw_stream_dequeue_buffer(impl->capture)) == NULL)
+			break;
+		if (in)
+			pw_stream_queue_buffer(impl->capture, in);
+		in = t;
+	}
+	if (in == NULL)
 		pw_log_debug("%p: out of capture buffers: %m", impl);
 
 	if ((out = pw_stream_dequeue_buffer(impl->playback)) == NULL)
@@ -816,7 +915,7 @@ static struct spa_pod *get_prop_info(struct graph *graph, struct spa_pod_builder
 	struct fc_port *p = &d->ports[port->p];
 	float def, min, max;
 	char name[512];
-	uint32_t rate = impl->rate ? impl->rate : 48000;
+	uint32_t rate = impl->rate ? impl->rate : DEFAULT_RATE;
 
 	if (p->hint & FC_HINT_SAMPLE_RATE) {
 		def = p->def * rate;
@@ -900,35 +999,48 @@ static struct spa_pod *get_props_param(struct graph *graph, struct spa_pod_build
 
 		spa_pod_builder_string(b, name);
 		if (p->hint & FC_HINT_BOOLEAN) {
-			spa_pod_builder_bool(b, port->control_data <= 0.0f ? false : true);
+			spa_pod_builder_bool(b, port->control_data[0] <= 0.0f ? false : true);
 		} else if (p->hint & FC_HINT_INTEGER) {
-			spa_pod_builder_int(b, port->control_data);
+			spa_pod_builder_int(b, port->control_data[0]);
 		} else {
-			spa_pod_builder_float(b, port->control_data);
+			spa_pod_builder_float(b, port->control_data[0]);
 		}
 	}
 	spa_pod_builder_pop(b, &f[1]);
 	return spa_pod_builder_pop(b, &f[0]);
 }
 
+static int port_set_control_value(struct port *port, float *value, uint32_t id)
+{
+	struct node *node = port->node;
+	struct descriptor *desc = node->desc;
+	float old;
+
+	old = port->control_data[id];
+	port->control_data[id] = value ? *value : desc->default_control[port->idx];
+	pw_log_info("control %d %d ('%s') from %f to %f", port->idx, id,
+			desc->desc->ports[port->p].name, old, port->control_data[id]);
+	node->control_changed = old != port->control_data[id];
+	return node->control_changed ? 1 : 0;
+}
+
 static int set_control_value(struct node *node, const char *name, float *value)
 {
-	struct descriptor *desc;
 	struct port *port;
-	float old;
+	int count = 0;
+	uint32_t i, n_hndl;
 
 	port = find_port(node, name, FC_PORT_INPUT | FC_PORT_CONTROL);
 	if (port == NULL)
 		return -ENOENT;
 
-	node = port->node;
-	desc = node->desc;
+	/* if we don't have any instances yet, set the first control value, we will
+	 * copy to other instances later */
+	n_hndl = SPA_MAX(1u, port->node->n_hndl);
+	for (i = 0; i < n_hndl; i++)
+		count += port_set_control_value(port, value, i);
 
-	old = port->control_data;
-	port->control_data = value ? *value : desc->default_control[port->idx];
-	pw_log_info("control %d ('%s') from %f to %f", port->idx, name, old, port->control_data);
-	node->control_changed = old != port->control_data;
-	return node->control_changed ? 1 : 0;
+	return count;
 }
 
 static int parse_params(struct graph *graph, const struct spa_pod *pod)
@@ -1006,32 +1118,138 @@ static void node_control_changed(struct node *node)
 	node->control_changed = false;
 }
 
-static void param_props_changed(struct impl *impl, const struct spa_pod *param)
+static void update_props_param(struct impl *impl)
+{
+	struct graph *graph = &impl->graph;
+	uint8_t buffer[1024];
+	struct spa_pod_dynamic_builder b;
+	const struct spa_pod *params[1];
+
+	spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+	params[0] = get_props_param(graph, &b.b);
+
+	pw_stream_update_params(impl->capture, params, 1);
+	spa_pod_dynamic_builder_clean(&b);
+}
+
+static int sync_volume(struct graph *graph, struct volume *vol)
+{
+	uint32_t i;
+	int res = 0;
+
+	if (vol->n_ports == 0)
+		return 0;
+	for (i = 0; i < vol->n_volumes; i++) {
+		uint32_t n_port = i % vol->n_ports, n_hndl;
+		struct port *p = vol->ports[n_port];
+		float v = vol->mute ? 0.0f : vol->volumes[i];
+		switch (vol->scale[n_port]) {
+		case SCALE_CUBIC:
+			v = cbrt(v);
+			break;
+		}
+		v = v * (vol->max[n_port] - vol->min[n_port]) + vol->min[n_port];
+
+		n_hndl = SPA_MAX(1u, p->node->n_hndl);
+		res += port_set_control_value(p, &v, i % n_hndl);
+	}
+	return res;
+}
+
+static void param_props_changed(struct impl *impl, const struct spa_pod *param,
+		bool capture)
 {
 	struct spa_pod_object *obj = (struct spa_pod_object *) param;
+	struct spa_pod_frame f[1];
 	const struct spa_pod_prop *prop;
 	struct graph *graph = &impl->graph;
 	int changed = 0;
+	char buf[1024];
+	struct spa_pod_dynamic_builder b;
+	struct volume *vol = capture ? &graph->capture_volume :
+		&graph->playback_volume;
+	bool do_volume = false;
+
+	spa_pod_dynamic_builder_init(&b, buf, sizeof(buf), 1024);
+	spa_pod_builder_push_object(&b.b, &f[0], SPA_TYPE_OBJECT_Props, SPA_PARAM_Props);
 
 	SPA_POD_OBJECT_FOREACH(obj, prop) {
-		if (prop->key == SPA_PROP_params)
+		switch (prop->key) {
+		case SPA_PROP_params:
 			changed += parse_params(graph, &prop->value);
+			spa_pod_builder_raw_padded(&b.b, prop, SPA_POD_PROP_SIZE(prop));
+			break;
+		case SPA_PROP_mute:
+		{
+			bool mute;
+			if (spa_pod_get_bool(&prop->value, &mute) == 0) {
+				if (vol->mute != mute) {
+					vol->mute = mute;
+					do_volume = true;
+				}
+			}
+			spa_pod_builder_raw_padded(&b.b, prop, SPA_POD_PROP_SIZE(prop));
+			break;
+		}
+		case SPA_PROP_channelVolumes:
+		{
+			uint32_t i, n_vols;
+			float vols[SPA_AUDIO_MAX_CHANNELS];
+
+			if ((n_vols = spa_pod_copy_array(&prop->value, SPA_TYPE_Float, vols,
+					SPA_AUDIO_MAX_CHANNELS)) > 0) {
+				if (vol->n_volumes != n_vols)
+					do_volume = true;
+				vol->n_volumes = n_vols;
+				for (i = 0; i < n_vols; i++) {
+					float v = vols[i];
+					if (v != vol->volumes[i]) {
+						vol->volumes[i] = v;
+						do_volume = true;
+					}
+				}
+			}
+			spa_pod_builder_raw_padded(&b.b, prop, SPA_POD_PROP_SIZE(prop));
+			break;
+		}
+		case SPA_PROP_softVolumes:
+		case SPA_PROP_softMute:
+			break;
+		default:
+			spa_pod_builder_raw_padded(&b.b, prop, SPA_POD_PROP_SIZE(prop));
+			break;
+		}
 	}
+	if (do_volume && vol->n_ports != 0) {
+		float soft_vols[SPA_AUDIO_MAX_CHANNELS];
+		uint32_t i;
+
+		for (i = 0; i < vol->n_volumes; i++)
+			soft_vols[i] = (vol->mute || vol->volumes[i] == 0.0f) ? 0.0f : 1.0f;
+
+		spa_pod_builder_prop(&b.b, SPA_PROP_softMute, 0);
+		spa_pod_builder_bool(&b.b, vol->mute);
+		spa_pod_builder_prop(&b.b, SPA_PROP_softVolumes, 0);
+		spa_pod_builder_array(&b.b, sizeof(float), SPA_TYPE_Float,
+				vol->n_volumes, soft_vols);
+		param = spa_pod_builder_pop(&b.b, &f[0]);
+
+		sync_volume(graph, vol);
+		pw_stream_set_param(capture ? impl->capture :
+				impl->playback, SPA_PARAM_Props, param);
+	}
+
+	spa_pod_dynamic_builder_clean(&b);
+
 	if (changed > 0) {
-		uint8_t buffer[1024];
-		struct spa_pod_dynamic_builder b;
-		const struct spa_pod *params[1];
 		struct node *node;
 
 		spa_list_for_each(node, &graph->node_list, link)
 			node_control_changed(node);
 
-		spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
-		params[0] = get_props_param(graph, &b.b);
-
-		pw_stream_update_params(impl->capture, params, 1);
-		spa_pod_dynamic_builder_clean(&b);
+		update_props_param(impl);
 	}
+
 }
 
 static void param_latency_changed(struct impl *impl, const struct spa_pod *param)
@@ -1041,7 +1259,7 @@ static void param_latency_changed(struct impl *impl, const struct spa_pod *param
 	struct spa_pod_builder b;
 	const struct spa_pod *params[1];
 
-	if (spa_latency_parse(param, &latency) < 0)
+	if (param == NULL || spa_latency_parse(param, &latency) < 0)
 		return;
 
 	spa_pod_builder_init(&b, buffer, sizeof(buffer));
@@ -1053,11 +1271,27 @@ static void param_latency_changed(struct impl *impl, const struct spa_pod *param
 		pw_stream_update_params(impl->playback, params, 1);
 }
 
+static void param_tag_changed(struct impl *impl, const struct spa_pod *param)
+{
+	struct spa_tag_info tag;
+	const struct spa_pod *params[1] = { param };
+	void *state = NULL;
+
+	if (param == 0 || spa_tag_parse(param, &tag, &state) < 0)
+		return;
+
+	if (tag.direction == SPA_DIRECTION_INPUT)
+		pw_stream_update_params(impl->capture, params, 1);
+	else
+		pw_stream_update_params(impl->playback, params, 1);
+}
+
 static void state_changed(void *data, enum pw_stream_state old,
 		enum pw_stream_state state, const char *error)
 {
 	struct impl *impl = data;
 	struct graph *graph = &impl->graph;
+	int res;
 
 	switch (state) {
 	case PW_STREAM_STATE_PAUSED:
@@ -1072,12 +1306,47 @@ static void state_changed(void *data, enum pw_stream_state old,
 	case PW_STREAM_STATE_ERROR:
 		pw_log_info("module %p: error: %s", impl, error);
 		break;
+	case PW_STREAM_STATE_STREAMING:
+	{
+		uint32_t target = impl->info.rate;
+		if (target == 0)
+			target = impl->position ?
+				impl->position->clock.target_rate.denom : DEFAULT_RATE;
+		if (target == 0) {
+			res = -EINVAL;
+			goto error;
+		}
+		if (impl->rate != target) {
+			impl->rate = target;
+			graph_cleanup(graph);
+			if ((res = graph_instantiate(graph)) < 0)
+				goto error;
+		}
+		break;
+	}
+	default:
+		break;
+	}
+	return;
+error:
+	pw_stream_set_error(impl->capture, res, "can't start graph: %s",
+			spa_strerror(res));
+}
+
+static void io_changed(void *data, uint32_t id, void *area, uint32_t size)
+{
+	struct impl *impl = data;
+	switch (id) {
+	case SPA_IO_Position:
+		impl->position = area;
+		break;
 	default:
 		break;
 	}
 }
 
-static void param_changed(void *data, uint32_t id, const struct spa_pod *param)
+static void param_changed(void *data, uint32_t id, const struct spa_pod *param,
+		bool capture)
 {
 	struct impl *impl = data;
 	struct graph *graph = &impl->graph;
@@ -1085,45 +1354,56 @@ static void param_changed(void *data, uint32_t id, const struct spa_pod *param)
 
 	switch (id) {
 	case SPA_PARAM_Format:
+	{
+		struct spa_audio_info_raw info;
+		spa_zero(info);
 		if (param == NULL) {
 			graph_cleanup(graph);
+			impl->rate = 0;
 		} else {
-			struct spa_audio_info_raw info;
-			spa_zero(info);
 			if ((res = spa_format_audio_raw_parse(param, &info)) < 0)
 				goto error;
-			if (info.rate == 0) {
-				res = -EINVAL;
-				goto error;
-			}
-			impl->rate = info.rate;
-			if ((res = graph_instantiate(graph)) < 0)
-				goto error;
 		}
+		impl->info = info;
 		break;
+	}
 	case SPA_PARAM_Props:
 		if (param != NULL)
-			param_props_changed(impl, param);
+			param_props_changed(impl, param, capture);
 		break;
 	case SPA_PARAM_Latency:
 		param_latency_changed(impl, param);
 		break;
+	case SPA_PARAM_Tag:
+		param_tag_changed(impl, param);
+		break;
 	}
 	return;
 
 error:
-	pw_stream_set_error(impl->capture, res, "can't start graph: %s",
-			spa_strerror(res));
+	pw_stream_set_error(capture ? impl->capture : impl->playback,
+			res, "can't start graph: %s", spa_strerror(res));
+}
+
+static void capture_param_changed(void *data, uint32_t id, const struct spa_pod *param)
+{
+	param_changed(data, id, param, true);
 }
 
 static const struct pw_stream_events in_stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = capture_destroy,
 	.process = capture_process,
+	.io_changed = io_changed,
 	.state_changed = state_changed,
-	.param_changed = param_changed
+	.param_changed = capture_param_changed
 };
 
+static void playback_param_changed(void *data, uint32_t id, const struct spa_pod *param)
+{
+	param_changed(data, id, param, false);
+}
+
 static void playback_destroy(void *d)
 {
 	struct impl *impl = d;
@@ -1135,8 +1415,9 @@ static const struct pw_stream_events out_stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = playback_destroy,
 	.process = playback_process,
+	.io_changed = io_changed,
 	.state_changed = state_changed,
-	.param_changed = param_changed
+	.param_changed = playback_param_changed,
 };
 
 static int setup_streams(struct impl *impl)
@@ -1208,7 +1489,8 @@ static int setup_streams(struct impl *impl)
 			PW_ID_ANY,
 			PW_STREAM_FLAG_AUTOCONNECT |
 			PW_STREAM_FLAG_MAP_BUFFERS |
-			PW_STREAM_FLAG_RT_PROCESS,
+			PW_STREAM_FLAG_RT_PROCESS |
+			PW_STREAM_FLAG_ASYNC,
 			params, n_params);
 
 	spa_pod_dynamic_builder_clean(&b);
@@ -1258,12 +1540,92 @@ static void plugin_unref(struct plugin *hndl)
 	free(hndl);
 }
 
+
+static struct plugin_func *add_plugin_func(struct impl *impl, const char *type,
+		fc_plugin_load_func *func, void *hndl)
+{
+	struct plugin_func *pl;
+
+	pl = calloc(1, sizeof(*pl));
+	if (pl == NULL)
+		return NULL;
+
+	snprintf(pl->type, sizeof(pl->type), "%s", type);
+	pl->func = func;
+	pl->hndl = hndl;
+	spa_list_append(&impl->plugin_func_list, &pl->link);
+	return pl;
+}
+
+static void free_plugin_func(struct plugin_func *pl)
+{
+	spa_list_remove(&pl->link);
+	if (pl->hndl)
+		dlclose(pl->hndl);
+	free(pl);
+}
+
+static fc_plugin_load_func *find_plugin_func(struct impl *impl, const char *type)
+{
+	fc_plugin_load_func *func = NULL;
+	void *hndl = NULL;
+	int res;
+	struct plugin_func *pl;
+	char module[PATH_MAX];
+	const char *module_dir;
+	const char *state = NULL, *p;
+	size_t len;
+
+	spa_list_for_each(pl, &impl->plugin_func_list, link) {
+		if (spa_streq(pl->type, type))
+			return pl->func;
+	}
+	module_dir = getenv("PIPEWIRE_MODULE_DIR");
+	if (module_dir == NULL)
+		module_dir = MODULEDIR;
+	pw_log_debug("moduledir set to: %s", module_dir);
+
+	while ((p = pw_split_walk(module_dir, ":", &len, &state))) {
+		if ((res = spa_scnprintf(module, sizeof(module),
+				"%.*s/libpipewire-module-filter-chain-%s.so",
+						(int)len, p, type)) <= 0)
+			continue;
+
+		hndl = dlopen(module, RTLD_NOW | RTLD_LOCAL);
+		if (hndl != NULL)
+			break;
+
+		pw_log_debug("open plugin module %s failed: %s", module, dlerror());
+	}
+	if (hndl == NULL) {
+		errno = ENOENT;
+		return NULL;
+	}
+	func = dlsym(hndl, FC_PLUGIN_LOAD_FUNC);
+	if (func != NULL) {
+		pw_log_info("opened plugin module %s", module);
+		pl = add_plugin_func(impl, type, func, hndl);
+		if (pl == NULL)
+			goto error_close;
+	} else {
+		errno = ENOSYS;
+		pw_log_error("%s is not a filter chain plugin: %m", module);
+		goto error_close;
+	}
+	return func;
+
+error_close:
+	dlclose(hndl);
+	return NULL;
+}
+
 static struct plugin *plugin_load(struct impl *impl, const char *type, const char *path)
 {
 	struct fc_plugin *pl = NULL;
 	struct plugin *hndl;
 	const struct spa_support *support;
 	uint32_t n_support;
+	fc_plugin_load_func *plugin_func;
 
 	spa_list_for_each(hndl, &impl->plugin_list, link) {
 		if (spa_streq(hndl->type, type) &&
@@ -1274,27 +1636,12 @@ static struct plugin *plugin_load(struct impl *impl, const char *type, const cha
 	}
 	support = pw_context_get_support(impl->context, &n_support);
 
-	if (spa_streq(type, "builtin")) {
-		pl = load_builtin_plugin(support, n_support, &impl->dsp, path, NULL);
-	}
-	else if (spa_streq(type, "sofa")) {
-		pl = load_sofa_plugin(support, n_support, &impl->dsp, path, NULL);
-	}
-	else if (spa_streq(type, "ladspa")) {
-		pl = load_ladspa_plugin(support, n_support, &impl->dsp, path, NULL);
-	}
-	else if (spa_streq(type, "lv2")) {
-#ifdef HAVE_LILV
-		pl = load_lv2_plugin(support, n_support, &impl->dsp, path, NULL);
-#else
-		pw_log_error("filter-chain is compiled without lv2 support");
+	plugin_func = find_plugin_func(impl, type);
+	if (plugin_func == NULL) {
+		pw_log_error("can't load plugin type '%s': %m", type);
 		pl = NULL;
-		errno = ENOTSUP;
-#endif
 	} else {
-		pw_log_error("invalid plugin type '%s'", type);
-		pl = NULL;
-		errno = EINVAL;
+		pl = plugin_func(support, n_support, &impl->dsp, path, NULL);
 	}
 	if (pl == NULL)
 		goto exit;
@@ -1307,7 +1654,7 @@ static struct plugin *plugin_load(struct impl *impl, const char *type, const cha
 	snprintf(hndl->type, sizeof(hndl->type), "%s", type);
 	snprintf(hndl->path, sizeof(hndl->path), "%s", path);
 
-	pw_log_info("successfully opened '%s'", path);
+	pw_log_info("successfully opened '%s':'%s'", type, path);
 
 	hndl->plugin = pl;
 
@@ -1513,7 +1860,7 @@ static int parse_link(struct graph *graph, struct spa_json *json)
 	char output[256] = "";
 	char input[256] = "";
 	const char *val;
-	struct node *def_node;
+	struct node *def_in_node, *def_out_node;
 	struct port *in_port, *out_port;
 	struct link *link;
 
@@ -1538,16 +1885,25 @@ static int parse_link(struct graph *graph, struct spa_json *json)
 		else if (spa_json_next(json, &val) < 0)
 			break;
 	}
-	def_node = spa_list_first(&graph->node_list, struct node, link);
-	if ((out_port = find_port(def_node, output, FC_PORT_OUTPUT)) == NULL) {
-		pw_log_error("unknown output port %s", output);
-		return -ENOENT;
+	def_out_node = spa_list_first(&graph->node_list, struct node, link);
+	def_in_node = spa_list_last(&graph->node_list, struct node, link);
+
+	out_port = find_port(def_out_node, output, FC_PORT_OUTPUT);
+	in_port = find_port(def_in_node, input, FC_PORT_INPUT);
+
+	if (out_port == NULL && out_port == NULL) {
+		/* try control ports */
+		out_port = find_port(def_out_node, output, FC_PORT_OUTPUT | FC_PORT_CONTROL);
+		in_port = find_port(def_in_node, input, FC_PORT_INPUT | FC_PORT_CONTROL);
 	}
-	def_node = spa_list_last(&graph->node_list, struct node, link);
-	if ((in_port = find_port(def_node, input, FC_PORT_INPUT)) == NULL) {
-		pw_log_error("unknown input port %s", input);
+	if (in_port == NULL || out_port == NULL) {
+		if (out_port == NULL)
+			pw_log_error("unknown output port %s", output);
+		if (in_port == NULL)
+			pw_log_error("unknown input port %s", input);
 		return -ENOENT;
 	}
+
 	if (in_port->n_links > 0) {
 		pw_log_info("Can't have more than 1 link to %s, use a mixer", input);
 		return -ENOTSUP;
@@ -1588,6 +1944,91 @@ static void link_free(struct link *link)
 	free(link);
 }
 
+/**
+ * {
+ *   control = [name:][portname]
+ *   min = <float, defaukt 0.0>
+ *   max = <float, default 1.0>
+ *   scale = <string, default "linear", options "linear","cubic">
+ * }
+ */
+static int parse_volume(struct graph *graph, struct spa_json *json, bool capture)
+{
+	char key[256];
+	char control[256] = "";
+	char scale[64] = "linear";
+	float min = 0.0f, max = 1.0f;
+	const char *val;
+	struct node *def_control;
+	struct port *port;
+	struct volume *vol = capture ? &graph->capture_volume :
+		&graph->playback_volume;
+
+	if (spa_list_is_empty(&graph->node_list)) {
+		pw_log_error("can't set volume in graph without nodes");
+		return -EINVAL;
+	}
+	while (spa_json_get_string(json, key, sizeof(key)) > 0) {
+		if (spa_streq(key, "control")) {
+			if (spa_json_get_string(json, control, sizeof(control)) <= 0) {
+				pw_log_error("control expects a string");
+				return -EINVAL;
+			}
+		}
+		else if (spa_streq(key, "min")) {
+			if (spa_json_get_float(json, &min) <= 0) {
+				pw_log_error("min expects a float");
+				return -EINVAL;
+			}
+		}
+		else if (spa_streq(key, "max")) {
+			if (spa_json_get_float(json, &max) <= 0) {
+				pw_log_error("max expects a float");
+				return -EINVAL;
+			}
+		}
+		else if (spa_streq(key, "scale")) {
+			if (spa_json_get_string(json, scale, sizeof(scale)) <= 0) {
+				pw_log_error("scale expects a string");
+				return -EINVAL;
+			}
+		}
+		else if (spa_json_next(json, &val) < 0)
+			break;
+	}
+	if (capture)
+		def_control = spa_list_first(&graph->node_list, struct node, link);
+	else
+		def_control = spa_list_last(&graph->node_list, struct node, link);
+
+	port = find_port(def_control, control, FC_PORT_INPUT | FC_PORT_CONTROL);
+	if (port == NULL) {
+		pw_log_error("unknown control port %s", control);
+		return -ENOENT;
+	}
+	if (vol->n_ports >= SPA_AUDIO_MAX_CHANNELS) {
+		pw_log_error("too many volume controls");
+		return -ENOSPC;
+	}
+	if (spa_streq(scale, "linear")) {
+		vol->scale[vol->n_ports] = SCALE_LINEAR;
+	} else if (spa_streq(scale, "cubic")) {
+		vol->scale[vol->n_ports] = SCALE_CUBIC;
+	} else {
+		pw_log_error("Invalid scale value '%s', use one of linear or cubic", scale);
+		return -EINVAL;
+	}
+	pw_log_info("volume %d: \"%s:%s\" min:%f max:%f scale:%s", vol->n_ports, port->node->name,
+			port->node->desc->desc->ports[port->p].name, min, max, scale);
+
+	vol->ports[vol->n_ports] = port;
+	vol->min[vol->n_ports] = min;
+	vol->max[vol->n_ports] = max;
+	vol->n_ports++;
+
+	return 0;
+}
+
 /**
  * type = ladspa
  * name = rev
@@ -1700,7 +2141,7 @@ static int load_node(struct graph *graph, struct spa_json *json)
 		port->external = SPA_ID_INVALID;
 		port->p = desc->control[i];
 		spa_list_init(&port->link_list);
-		port->control_data = desc->default_control[i];
+		port->control_data[0] = desc->default_control[i];
 	}
 	for (i = 0; i < desc->n_notify; i++) {
 		struct port *port = &node->notify_port[i];
@@ -1729,6 +2170,7 @@ static void node_cleanup(struct node *node)
 	for (i = 0; i < node->n_hndl; i++) {
 		if (node->hndl[i] == NULL)
 			continue;
+		pw_log_info("cleanup %s %d", d->name, i);
 		if (d->deactivate)
 			d->deactivate(node->hndl[i]);
 		d->cleanup(node->hndl[i]);
@@ -1778,6 +2220,9 @@ static void node_free(struct node *node)
 static void graph_cleanup(struct graph *graph)
 {
 	struct node *node;
+	if (!graph->instantiated)
+		return;
+	graph->instantiated = false;
 	spa_list_for_each(node, &graph->node_list, link)
 		node_cleanup(node);
 }
@@ -1793,6 +2238,11 @@ static int graph_instantiate(struct graph *graph)
 	uint32_t i, j;
 	int res;
 
+	if (graph->instantiated)
+		return 0;
+
+	graph->instantiated = true;
+
 	spa_list_for_each(node, &graph->node_list, link) {
 		float *sd = silence_data, *dd = discard_data;
 
@@ -1805,8 +2255,9 @@ static int graph_instantiate(struct graph *graph)
 
 		for (i = 0; i < node->n_hndl; i++) {
 			pw_log_info("instantiate %s %d rate:%lu", d->name, i, impl->rate);
+			errno = EINVAL;
 			if ((node->hndl[i] = d->instantiate(d, impl->rate, i, node->config)) == NULL) {
-				pw_log_error("cannot create plugin instance: %m");
+				pw_log_error("cannot create plugin instance %d rate:%lu: %m", i, impl->rate);
 				res = -errno;
 				goto error;
 			}
@@ -1835,11 +2286,22 @@ static int graph_instantiate(struct graph *graph)
 			}
 			for (j = 0; j < desc->n_control; j++) {
 				port = &node->control_port[j];
-				d->connect_port(node->hndl[i], port->p, &port->control_data);
+				d->connect_port(node->hndl[i], port->p, &port->control_data[i]);
+
+				spa_list_for_each(link, &port->link_list, input_link) {
+					struct port *peer = link->output;
+					pw_log_info("connect control port %s[%d]:%s %p",
+							node->name, i, d->ports[port->p].name,
+							&peer->control_data[i]);
+					d->connect_port(node->hndl[i], port->p, &peer->control_data[i]);
+				}
 			}
 			for (j = 0; j < desc->n_notify; j++) {
 				port = &node->notify_port[j];
-				d->connect_port(node->hndl[i], port->p, &port->control_data);
+				pw_log_info("connect notify port %s[%d]:%s %p",
+						node->name, i, d->ports[port->p].name,
+						&port->control_data[i]);
+				d->connect_port(node->hndl[i], port->p, &port->control_data[i]);
 			}
 			if (d->activate)
 				d->activate(node->hndl[i]);
@@ -1847,12 +2309,29 @@ static int graph_instantiate(struct graph *graph)
 				d->control_changed(node->hndl[i]);
 		}
 	}
+	update_props_param(impl);
 	return 0;
 error:
 	graph_cleanup(graph);
 	return res;
 }
 
+/* any default values for the controls are set in the first instance
+ * of the control data. Duplicate this to the other instances now. */
+static void setup_node_controls(struct node *node)
+{
+	uint32_t i, j;
+	uint32_t n_hndl = node->n_hndl;
+	uint32_t n_ports = node->desc->n_control;
+	struct port *ports = node->control_port;
+
+	for (i = 0; i < n_ports; i++) {
+		struct port *port = &ports[i];
+		for (j = 1; j < n_hndl; j++)
+			port->control_data[j] = port->control_data[0];
+	}
+}
+
 static struct node *find_next_node(struct graph *graph)
 {
 	struct node *node;
@@ -1949,6 +2428,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 		desc = node->desc;
 		n_control += desc->n_control;
 		n_nodes++;
+		setup_node_controls(node);
 	}
 	graph->n_input = 0;
 	graph->input = calloc(n_input * 16 * n_hndl, sizeof(struct graph_port));
@@ -2137,6 +2617,7 @@ static int load_graph(struct graph *graph, struct pw_properties *props)
 {
 	struct spa_json it[3];
 	struct spa_json inputs, outputs, *pinputs = NULL, *poutputs = NULL;
+	struct spa_json cvolumes, pvolumes, *pcvolumes = NULL, *ppvolumes = NULL;
 	struct spa_json nodes, *pnodes = NULL, links, *plinks = NULL;
 	const char *json, *val;
 	char key[256];
@@ -2184,6 +2665,20 @@ static int load_graph(struct graph *graph, struct pw_properties *props)
 				return -EINVAL;
 			}
 			poutputs = &outputs;
+		}
+		else if (spa_streq("capture.volumes", key)) {
+			if (spa_json_enter_array(&it[1], &cvolumes) <= 0) {
+				pw_log_error("capture.volumes expects an array");
+				return -EINVAL;
+			}
+			pcvolumes = &cvolumes;
+		}
+		else if (spa_streq("playback.volumes", key)) {
+			if (spa_json_enter_array(&it[1], &pvolumes) <= 0) {
+				pw_log_error("playback.volumes expects an array");
+				return -EINVAL;
+			}
+			ppvolumes = &pvolumes;
 		} else if (spa_json_next(&it[1], &val) < 0)
 			break;
 	}
@@ -2201,6 +2696,18 @@ static int load_graph(struct graph *graph, struct pw_properties *props)
 				return res;
 		}
 	}
+	if (pcvolumes != NULL) {
+		while (spa_json_enter_object(pcvolumes, &it[2]) > 0) {
+			if ((res = parse_volume(graph, &it[2], true)) < 0)
+				return res;
+		}
+	}
+	if (ppvolumes != NULL) {
+		while (spa_json_enter_object(ppvolumes, &it[2]) > 0) {
+			if ((res = parse_volume(graph, &it[2], false)) < 0)
+				return res;
+		}
+	}
 	return setup_graph(graph, pinputs, poutputs);
 }
 
@@ -2253,6 +2760,8 @@ static const struct pw_proxy_events core_proxy_events = {
 
 static void impl_destroy(struct impl *impl)
 {
+	struct plugin_func *pl;
+
 	/* disconnect both streams before destroying any of them */
 	if (impl->capture)
 		pw_stream_disconnect(impl->capture);
@@ -2270,6 +2779,8 @@ static void impl_destroy(struct impl *impl)
 	pw_properties_free(impl->capture_props);
 	pw_properties_free(impl->playback_props);
 	graph_free(&impl->graph);
+	spa_list_consume(pl, &impl->plugin_func_list, link)
+		free_plugin_func(pl);
 	free(impl);
 }
 
@@ -2381,6 +2892,10 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	impl->graph.impl = impl;
 
 	spa_list_init(&impl->plugin_list);
+	spa_list_init(&impl->plugin_func_list);
+
+	add_plugin_func(impl, "builtin", load_builtin_plugin, NULL);
+	add_plugin_func(impl, "ladspa", load_ladspa_plugin, NULL);
 
 	support = pw_context_get_support(impl->context, &n_support);
 
@@ -2418,7 +2933,12 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	parse_audio_info(impl->capture_props, &impl->capture_info);
 	parse_audio_info(impl->playback_props, &impl->playback_info);
 
-	if (impl->capture_info.rate && !impl->playback_info.rate)
+	if (!impl->capture_info.rate && !impl->playback_info.rate) {
+		if (pw_properties_get(impl->playback_props, "resample.disable") == NULL)
+			pw_properties_set(impl->playback_props, "resample.disable", "true");
+		if (pw_properties_get(impl->capture_props, "resample.disable") == NULL)
+			pw_properties_set(impl->capture_props, "resample.disable", "true");
+	} else if (impl->capture_info.rate && !impl->playback_info.rate)
 		impl->playback_info.rate = impl->capture_info.rate;
 	else if (impl->playback_info.rate && !impl->capture_info.rate)
 		impl->capture_info.rate = !impl->playback_info.rate;
