@@ -1,33 +1,13 @@
-/* PipeWire
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include "config.h"
 
 #include <unistd.h>
 #include <limits.h>
 #include <stdio.h>
-#if !defined(__FreeBSD__) && !defined(__MidnightBSD__)
+#if !defined(__FreeBSD__) && !defined(__MidnightBSD__) && !defined(__GNU__)
 #include <sys/prctl.h>
 #endif
 #include <pwd.h>
@@ -45,8 +25,10 @@
 #include <spa/support/cpu.h>
 #include <spa/support/i18n.h>
 
+#include <pipewire/cleanup.h>
 #include "pipewire.h"
 #include "private.h"
+#include "i18n.h"
 
 #define MAX_SUPPORT	32
 
@@ -517,9 +499,7 @@ static char *
 parse_pw_debug_env(void)
 {
 	const char *str;
-	char **tokens;
 	int n_tokens;
-	size_t slen;
 	char json[1024] = {0};
 	char *pos = json;
 	char *end = pos + sizeof(json) - 1;
@@ -527,26 +507,25 @@ parse_pw_debug_env(void)
 
 	str = getenv("PIPEWIRE_DEBUG");
 
-	if (!str || (slen = strlen(str)) == 0)
+	if (!str || !*str)
 		return NULL;
 
-	/* String format is PIPEWIRE_DEBUG=<glob>:<level>[,<glob>:<level>,...],
+	/* String format is PIPEWIRE_DEBUG=[<glob>:]<level>,...,
 	 * converted into [{ conn.* = 0}, {glob = level}, {glob = level}, ....] ,
 	 * with the connection namespace disabled by default.
 	 */
 	pos += spa_scnprintf(pos, end - pos, "[ { conn.* = %d },", SPA_LOG_LEVEL_NONE);
 
-	tokens = pw_split_strv(str, ",", INT_MAX, &n_tokens);
+	spa_auto(pw_strv) tokens = pw_split_strv(str, ",", INT_MAX, &n_tokens);
 	if (n_tokens > 0) {
 		int i;
 		for (i = 0; i < n_tokens; i++) {
 			int n_tok;
-			char **tok;
-			char *pattern;
+			char *tok[2];
 
-			tok = pw_split_strv(tokens[i], ":", 2, &n_tok);
+			n_tok = pw_split_ip(tokens[i], ":", SPA_N_ELEMENTS(tok), tok);
 			if (n_tok == 2 && parse_log_level(tok[1], &lvl)) {
-				pattern = tok[0];
+				char *pattern = tok[0];
 				pos += spa_scnprintf(pos, end - pos, "{ %s = %d },",
 						     pattern, lvl);
 			} else if (n_tok == 1 && parse_log_level(tok[0], &lvl)) {
@@ -555,11 +534,9 @@ parse_pw_debug_env(void)
 				pw_log_warn("Ignoring invalid format in PIPEWIRE_DEBUG: '%s'",
 						tokens[i]);
 			}
-
-			pw_free_strv(tok);
 		}
 	}
-	pw_free_strv(tokens);
+
 	pos += spa_scnprintf(pos, end - pos, "]");
 	return strdup(json);
 }
@@ -589,6 +566,8 @@ void pw_init(int *argc, char **argv[])
 	if (support->init_count > 0)
 		goto done;
 
+	pw_random_init();
+
 	pthread_mutex_lock(&support_lock);
 	support->in_valgrind = RUNNING_ON_VALGRIND;
 
@@ -619,8 +598,11 @@ void pw_init(int *argc, char **argv[])
 		char *patterns = NULL;
 
 		n_items = 0;
-		if (!support->no_color)
-			items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_LOG_COLORS, "true");
+		if (!support->no_color) {
+			if ((str = getenv("PIPEWIRE_LOG_COLOR")) == NULL)
+				str = "true";
+			items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_LOG_COLORS, str);
+		}
 		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_LOG_TIMESTAMP, "true");
 		if ((str = getenv("PIPEWIRE_LOG_LINE")) == NULL || spa_atob(str))
 			items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_LOG_LINE, "true");
@@ -738,7 +720,7 @@ static void init_prgname(void)
 	static char name[PATH_MAX];
 
 	spa_memzero(name, sizeof(name));
-#if defined(__linux__) || defined(__FreeBSD_kernel__) || defined(__MidnightBSD_kernel__)
+#if defined(__linux__) || defined(__FreeBSD_kernel__) || defined(__MidnightBSD_kernel__) || defined(__GNU__)
 	{
 		if (readlink("/proc/self/exe", name, sizeof(name)-1) > 0) {
 			prgname = strrchr(name, '/') + 1;
@@ -756,7 +738,7 @@ static void init_prgname(void)
 		}
 	}
 #endif
-#if !defined(__FreeBSD__) && !defined(__MidnightBSD__)
+#if !defined(__FreeBSD__) && !defined(__MidnightBSD__) && !defined(__GNU__)
 	{
 		if (prctl(PR_GET_NAME, (unsigned long) name, 0, 0, 0) == 0) {
 			prgname = name;
@@ -859,6 +841,12 @@ const char* pw_get_library_version(void)
 	return pw_get_headers_version();
 }
 
+SPA_EXPORT
+bool pw_check_library_version(int major, int minor, int micro)
+{
+	return PW_CHECK_VERSION(major, minor, micro);
+}
+
 static const struct spa_type_info type_info[] = {
 	{ SPA_ID_INVALID, SPA_ID_INVALID, "spa_types", spa_types },
 	{ 0, 0, NULL, NULL },
